// TexVertBuf: use a floating-point surface to store the
// transformed vertex position data and reuse it when rendering

// written by Christian Mendl, released under the GPL

#include <d3d9.h>
#include <d3dx9.h>
#include <dxerr9.h>
#include <tchar.h>
#include <mmsystem.h>
#include "vs.h"


LPDIRECT3D9 pD3D;
LPDIRECT3DDEVICE9 pd3dDevice;

LPDIRECT3DVERTEXSHADER9 VSRnd, VSCalc;
LPDIRECT3DVERTEXDECLARATION9 VDeclRnd, VDeclCalc;

LPDIRECT3DPIXELSHADER9 PSRnd, PSCalc;

// store the transformed position data and diffuse color;
// 2 vertex buffers, because VBufRndPos has to be dynamic and VBufRndDiff
// can remain static
LPDIRECT3DVERTEXBUFFER9 VBufRndPos, VBufRndDiff;

// untransformed position data
LPDIRECT3DVERTEXBUFFER9 VBufCalc;

// number of vertices
const unsigned int NumVert = 4;

// the floating-point render target
LPDIRECT3DSURFACE9 TexVertBuf;


struct CVertCalc
{
	// untransformed vertex position
	D3DVECTOR Pos;

	// the "id" used for mapping the vertex to
	// the appropriate pixel of TexVertBuf
	float id;
};


// concatenated view and projection matrix
D3DXMATRIX MtrViewProj;


// convert the index in the vertex buffer of a
// given vertex to the appropriate "id" used to
// pin it onto our floating-point render target
// l: number of vertices in the buffer
inline float i2id(const int i, const float l)
{
	return (2.f*(float)i+1.f)/l - 1.f;
}


// funtion called once at program startup
HRESULT setUp(const HWND& hWnd)
{
	HRESULT hr;


	if ((pD3D = Direct3DCreate9(D3D_SDK_VERSION)) == NULL)
		return E_FAIL;

	D3DPRESENT_PARAMETERS d3dpp;
	ZeroMemory(&d3dpp, sizeof(d3dpp));
	d3dpp.BackBufferFormat = D3DFMT_UNKNOWN;
	d3dpp.SwapEffect = D3DSWAPEFFECT_DISCARD;
	d3dpp.Windowed = true;

	if (FAILED(hr = pD3D->CreateDevice(D3DADAPTER_DEFAULT, D3DDEVTYPE_HAL, hWnd,
		D3DCREATE_HARDWARE_VERTEXPROCESSING, &d3dpp, &pd3dDevice))) return hr;


	// create the floating-point render target, 128 bits per pixel
	if (FAILED(hr = pd3dDevice->CreateRenderTarget(NumVert, 1, D3DFMT_A32B32G32R32F, D3DMULTISAMPLE_NONE, 0,
		true, &TexVertBuf, NULL))) return hr;


	// vertex declarations

	D3DVERTEXELEMENT9 declRnd[] =
	{
		{ 1, 0, D3DDECLTYPE_FLOAT3, D3DDECLMETHOD_DEFAULT, D3DDECLUSAGE_POSITION, 0 },
		{ 2, 0, D3DDECLTYPE_D3DCOLOR, D3DDECLMETHOD_DEFAULT, D3DDECLUSAGE_COLOR, 0 },
		D3DDECL_END()
	};
	if (FAILED(hr = pd3dDevice->CreateVertexDeclaration(declRnd, &VDeclRnd))) return hr;

	D3DVERTEXELEMENT9 declCalc[] =
	{
		{ 0,  0, D3DDECLTYPE_FLOAT3, D3DDECLMETHOD_DEFAULT, D3DDECLUSAGE_POSITION, 0 },
		{ 0, 12, D3DDECLTYPE_FLOAT1, D3DDECLMETHOD_DEFAULT, D3DDECLUSAGE_TEXCOORD, 0 },	// here the "id" is passed
		D3DDECL_END()
	};
	if (FAILED(hr = pd3dDevice->CreateVertexDeclaration(declCalc, &VDeclCalc))) return hr;


	// load and assemble vertex and pixel shaders

	LPD3DXBUFFER Code;

	if (FAILED(hr = D3DXAssembleShaderFromFile(_T("rnd.vsh"), NULL, NULL, 0, &Code, NULL))) return hr;
	if (FAILED(hr = pd3dDevice->CreateVertexShader((DWORD *) Code->GetBufferPointer(), &VSRnd))) return hr;

	if (FAILED(hr = D3DXAssembleShaderFromFile(_T("calc.vsh"), NULL, NULL, 0, &Code, NULL))) return hr;
	if (FAILED(hr = pd3dDevice->CreateVertexShader((DWORD *) Code->GetBufferPointer(), &VSCalc))) return hr;

	if (FAILED(hr = D3DXAssembleShaderFromFile(_T("rnd.psh"), NULL, NULL, 0, &Code, NULL))) return hr;
	if (FAILED(hr = pd3dDevice->CreatePixelShader((DWORD *) Code->GetBufferPointer(), &PSRnd))) return hr;

	if (FAILED(hr = D3DXAssembleShaderFromFile(_T("calc.psh"), NULL, NULL, 0, &Code, NULL))) return hr;
	if (FAILED(hr = pd3dDevice->CreatePixelShader((DWORD *) Code->GetBufferPointer(), &PSCalc))) return hr;


	// vertex buffers

	if (FAILED(hr = pd3dDevice->CreateVertexBuffer(NumVert * sizeof(D3DVECTOR), D3DUSAGE_DYNAMIC | D3DUSAGE_WRITEONLY,
		0, D3DPOOL_DEFAULT, &VBufRndPos, NULL))) return hr;

	if (FAILED(hr = pd3dDevice->CreateVertexBuffer(NumVert * sizeof(D3DCOLOR), 0,
		0, D3DPOOL_DEFAULT, &VBufRndDiff, NULL))) return hr;

	D3DCOLOR *VertRndDiff;
	if (FAILED(hr = VBufRndDiff->Lock(0, 0, (void **) &VertRndDiff, 0))) return hr;

	VertRndDiff[0] = 0xff00ff00;
	VertRndDiff[1] = 0xff0000ff;
	VertRndDiff[2] = 0xffff0000;
	VertRndDiff[3] = 0xffffffff;

	VBufRndDiff->Unlock();

	pd3dDevice->SetStreamSource(1, VBufRndPos, 0, sizeof(D3DVECTOR));
	pd3dDevice->SetStreamSource(2, VBufRndDiff, 0, sizeof(D3DCOLOR));


	if (FAILED(hr = pd3dDevice->CreateVertexBuffer(NumVert * sizeof(CVertCalc), D3DUSAGE_POINTS,
		0, D3DPOOL_DEFAULT, &VBufCalc, NULL))) return hr;

	// a quad

	CVertCalc *VertCalc;
	if (FAILED(hr = VBufCalc->Lock(0, 0, (void **) &VertCalc, 0))) return hr;

	VertCalc[0].Pos.x = -1;
	VertCalc[0].Pos.y = 1;
	VertCalc[0].Pos.z = 0;
	VertCalc[0].id = i2id(0, NumVert);

	VertCalc[1].Pos.x = 1;
	VertCalc[1].Pos.y = 1;
	VertCalc[1].Pos.z = 0;
	VertCalc[1].id = i2id(1, NumVert);

	VertCalc[2].Pos.x = -1;
	VertCalc[2].Pos.y = -1;
	VertCalc[2].Pos.z = 0;
	VertCalc[2].id = i2id(2, NumVert);

	VertCalc[3].Pos.x = 1;
	VertCalc[3].Pos.y = -1;
	VertCalc[3].Pos.z = 0;
	VertCalc[3].id = i2id(3, NumVert);

	VBufCalc->Unlock();

	pd3dDevice->SetStreamSource(0, VBufCalc, 0, sizeof(CVertCalc));


	// view and projection matrices

	D3DXMATRIX MtrView;
	D3DXVECTOR3 Eye(2.f, 1.f, -4.f);
	D3DXVECTOR3 At(0.f, 0.f, 0.f);
	D3DXVECTOR3 Up(0.f, 1.f, 0.f);
	D3DXMatrixLookAtLH(&MtrView, &Eye, &At, &Up);

	D3DXMATRIX MtrProj;
	D3DXMatrixPerspectiveFovLH(&MtrProj, D3DX_PI/4.f, 1.f, 1.f, 32.f);

	MtrViewProj = MtrView * MtrProj;


	return S_OK;
}


void cleanup()
{
	TexVertBuf->Release();

	VBufCalc->Release();
	VBufRndDiff->Release();
	VBufRndPos->Release();

	PSCalc->Release();
	PSRnd->Release();

	VDeclCalc->Release();
	VDeclRnd->Release();
	VSCalc->Release();
	VSRnd->Release();

	pd3dDevice->Release();
	pD3D->Release();
}


void moveFrame()
{
	// animate the scene

	D3DXMATRIX MtrWorld;
	D3DXMatrixRotationX(&MtrWorld, sin(timeGetTime()/1024.f));

	D3DMATRIX MtrWorldViewProj = MtrWorld * MtrViewProj;
	pd3dDevice->SetVertexShaderConstantF(MTR_WVP, (float *) &MtrWorldViewProj, 4);


	// first rendering pass: transform position data and store it in TexVertBuf

	pd3dDevice->SetVertexDeclaration(VDeclCalc);
	pd3dDevice->SetVertexShader(VSCalc);
	pd3dDevice->SetPixelShader(PSCalc);

	pd3dDevice->SetRenderTarget(0, TexVertBuf);

	pd3dDevice->BeginScene();

	pd3dDevice->DrawPrimitive(D3DPT_POINTLIST, 0, NumVert);

	pd3dDevice->EndScene();

	// don't call Present(...) since the surface
	// isn't supposed to be visible on screen

	// restore render target
	LPDIRECT3DSURFACE9 BackBuffer;
	pd3dDevice->GetBackBuffer(0, 0, D3DBACKBUFFER_TYPE_MONO, &BackBuffer);
	pd3dDevice->SetRenderTarget(0, BackBuffer);


	// copy data 

	D3DLOCKED_RECT RectCalcBuf;
	TexVertBuf->LockRect(&RectCalcBuf, NULL, D3DLOCK_READONLY);

	D3DVECTOR *VertRndPos;
	VBufRndPos->Lock(0, 0, (void **) &VertRndPos, D3DLOCK_DISCARD);

	// red maps to x, green to y and blue to z;
	// alpha is ignored (standard w value: 1)
	for (unsigned int i = 0; i < NumVert; i++)
		memcpy(&VertRndPos[i], (BYTE *)RectCalcBuf.pBits + i*16, sizeof(D3DVECTOR));

	VBufRndPos->Unlock();
	TexVertBuf->UnlockRect();
}


void render()
{
	pd3dDevice->Clear(0, NULL, D3DCLEAR_TARGET, 0xff000000, 1.0f, 0);

	pd3dDevice->BeginScene();

	pd3dDevice->SetVertexDeclaration(VDeclRnd);
	pd3dDevice->SetVertexShader(VSRnd);
	pd3dDevice->SetPixelShader(PSRnd);

	// draw a quad
	pd3dDevice->DrawPrimitive(D3DPT_TRIANGLESTRIP, 0, 2);

	pd3dDevice->EndScene();

	pd3dDevice->Present(NULL, NULL, NULL, NULL);
}


// Windows API stuff

LRESULT WINAPI msgHandler(HWND hWnd, UINT Message, WPARAM wParam, LPARAM lParam)
{
	switch(Message)
	{
		case WM_DESTROY:
			PostQuitMessage(0);
			return 0;

		case WM_PAINT:
			moveFrame();
			render();
			ValidateRect(hWnd, NULL);
			return 0;
	}

	return DefWindowProc(hWnd, Message, wParam, lParam);
}


int WINAPI WinMain(HINSTANCE hInst, HINSTANCE, LPSTR, int)
{
	WNDCLASSEX wc = { sizeof(WNDCLASSEX), CS_CLASSDC, msgHandler, 0, 0,
		hInst, NULL, NULL, NULL, NULL, _T("TexVertBuf"), NULL };

	RegisterClassEx(&wc);

	HWND hWnd = CreateWindow(_T("TexVertBuf"), _T("TexVertBuf"), WS_OVERLAPPEDWINDOW, 100, 100, 600, 600,
				GetDesktopWindow(), NULL, wc.hInstance, NULL);

	HRESULT hr;
	if (FAILED(hr = setUp(hWnd)))
	{
		char buf[128];
		_tcscpy(buf, _T("An error occured during data setup: "));
		_tcscat(buf, DXGetErrorDescription9(hr));
		_tcscat(buf, _T("\nThe application has to be terminated."));
		MessageBox(NULL, buf, _T("TexVertBuf"), MB_OK);

		return hr;
	}

	ShowWindow(hWnd, SW_SHOWDEFAULT);
	UpdateWindow(hWnd);

	MSG Message;
	do
	{
		if (PeekMessage(&Message, NULL, 0, 0, PM_REMOVE))
		{
			TranslateMessage(&Message);
			DispatchMessage(&Message);
		}
		else
		{
			moveFrame();
			render();
		}
	} while (Message.message != WM_QUIT);

	cleanup();
	UnregisterClass(_T("TexVertBuf"), wc.hInstance);

	return 0;
}